/*
   ┏━━━┓╱╱╱╱╱┏┓╱╱┏┓╱╱╱╱╱┏┓╱╱
   ┃┏━┓┃╱╱╱╱╱┃┃╱╱┃┃╱╱╱╱╱┃┃╱╱╱
   ┃┗━┛┣┓┏┳━━┫┃┏┓┃┃╱╱┏━━┫┗━┳━━┓
   ┃┏━━┫┃┃┃┏┓┃┗┛┃┃┃╱┏┫┏┓┃┏┓┃━━┫
   ┃┃╱╱┃┗┛┃┃┃┃┏┓┃┃┗━┛┃┏┓┃┗┛┣━━┃
   ┗┛╱╱┗━━┻┛┗┻┛┗┛┗━━━┻┛┗┻━━┻━━┛
    ━━━━━━━━━━━━━━━━━━━━━━━━━━

    Copyright (c) 2023 Punk Labs LLC

    This section is part of OneTrick URCHIN

    OneTrick URCHIN is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    OneTrick URCHIN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    OneTrick URCHIN.  If not, see <http://www.gnu.org/licenses/>.
*/

use std::cell::RefCell;
//use std::rc::Rc;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;

#[cfg(feature = "profile")]
use std::sync::atomic::AtomicI32;

use nih_plug::prelude::*;

// #[cfg(feature = "egui")]
// use egui_extras::image::RetainedImage;

#[cfg(feature = "egui")]
use nih_plug_egui::{
    create_egui_editor,
    egui::{
        epaint::{
            Shadow,
        },
        style,
        Margin,
        vec2,
        //Align,
        Align2,
        Area,
        CentralPanel,
        Color32,
        Context,
        FontDefinitions,
        FontFamily,
        FontId,
        Frame,
        Id,
        //LayerId,
        //Layout,
        Order,
        Pos2,
        Rect,
        Rounding,
        Sense,
        Stroke,
        Vec2,
    },
    EguiState,
};

mod dsp;
use onetrick::prelude::*;

//256 is 187.5Hz at 48kHz
//512 is 94Hz at 48kHz
//1024 is 47Hz at 48kHz
const MAX_BLOCK_SIZE: usize = 4096;
const NUM_CHANNELS: u32 = 2;
#[cfg(feature = "multi_out")]
const AUX_BUS_COUNT: usize = 7;

// NOTE: Should probably use an array of &dyn OneTrickGeneral, but couldn't make it work
macro_rules! for_each_voice_dsp {
    ($this:ident, $expression:expr) => {{
        $expression(&mut *$this.dsp_kick.borrow_mut());
        $expression(&mut *$this.dsp_snare.borrow_mut());
        $expression(&mut *$this.dsp_lowtom.borrow_mut());
        $expression(&mut *$this.dsp_midtom.borrow_mut());
        $expression(&mut *$this.dsp_hitom.borrow_mut());
        $expression(&mut *$this.dsp_hihat.borrow_mut());
        $expression(&mut *$this.dsp_cymbal.borrow_mut());
    }};
}

#[cfg(feature = "multi_out")]
macro_rules! for_each_voice_dsp_grouped {
    ($this:ident, $expression:expr) => {{
        $expression(&[&mut *$this.dsp_kick.borrow_mut()]);
        $expression(&[&mut *$this.dsp_snare.borrow_mut()]);
        $expression(&[&mut *$this.dsp_lowtom.borrow_mut()]);
        $expression(&[&mut *$this.dsp_midtom.borrow_mut()]);
        $expression(&[&mut *$this.dsp_hitom.borrow_mut()]);
        $expression(&[&mut *$this.dsp_hihat.borrow_mut()]);
        $expression(&[&mut *$this.dsp_cymbal.borrow_mut()]);
    }};
}
macro_rules! for_each_dsp {
    ($this:ident, $expression:expr) => {{
        for_each_voice_dsp!($this, $expression);
        $expression(&mut *$this.dsp_output.borrow_mut());
    }};
}

/// Main Struct for OneTrick Urchin
pub struct OneTrickUrchin {
    params: Arc<OneTrickPluginParams>,
    //processor: FixedProcessor,

    accum_buffer: ResizableBuffer,

    dsp_output: RefCell<OneTrickDSP<dsp::modules::output::DSP_Output>>,
    dsp_kick: RefCell<OneTrickDSP<dsp::modules::drum::DSP_Drum>>,
    dsp_snare: RefCell<OneTrickDSP<dsp::modules::snare::DSP_Snare>>,
    dsp_lowtom: RefCell<OneTrickDSP<dsp::modules::drum::DSP_Drum>>,
    dsp_midtom: RefCell<OneTrickDSP<dsp::modules::drum::DSP_Drum>>,
    dsp_hitom: RefCell<OneTrickDSP<dsp::modules::drum::DSP_Drum>>,
    dsp_hihat: RefCell<OneTrickDSP<dsp::modules::cymbal::DSP_Cymbal>>,
    dsp_cymbal: RefCell<OneTrickDSP<dsp::modules::cymbal::DSP_Cymbal>>,
    //dsp: [dyn OneTrickDSPGeneral],
    param_global_sensitivity: usize,
    param_global_transpose: usize,
    param_hihat_closed: i32,
    param_snare_rimshot: i32,

    kick_indicator: Arc<AtomicBool>,
    snare_indicator: Arc<AtomicBool>,
    lowtom_indicator: Arc<AtomicBool>,
    midtom_indicator: Arc<AtomicBool>,
    hitom_indicator: Arc<AtomicBool>,
    hihat_indicator: Arc<AtomicBool>,
    cymbal_indicator: Arc<AtomicBool>,

    #[cfg(feature = "profile")]
    dsp_cost: Arc<AtomicI32>,
    sample_rate: usize,
}

impl Default for OneTrickUrchin {
    fn default() -> Self {
        /*
        {
            use faust_types::FaustDsp;
            let explosion = Box::new(dsp::modules::DSP_Output::new());
        }
        */

        /* Generate Params?
            let params = OneTrickPluginParams::default();
            // Maybe use indices to reference subgroups?... Group names can be used by NIH-Plug
            params.append_dsp(&dsp_kick, 0, "Kick");
            params.append_dsp(&dsp_snare, 1, "Snare");
            ...
            params.append_dsp(&dsp_output, 11, "");

            //Later when we need to update params:
            params.apply_to_dsp(&mut dsp_kick, 0);
            params.apply_to_dsp(&mut dsp_snare, 1);
            ...
            params.apply_to_dsp(&mut dsp_output, 11);
        */

        let mut params = OneTrickPluginParams::default();
        #[cfg(feature = "egui")]
        {
            params.editor_state = EguiState::from_size(1063, 725);
        }

        let mut dsp_output = OneTrickDSP::new();
        let mut dsp_kick = OneTrickDSP::new();
        let mut dsp_snare = OneTrickDSP::new();
        let mut dsp_lowtom = OneTrickDSP::new();
        let mut dsp_midtom = OneTrickDSP::new();
        let mut dsp_hitom = OneTrickDSP::new();
        let mut dsp_hihat = OneTrickDSP::new();
        let mut dsp_cymbal = OneTrickDSP::new();

        // ==== OVERRIDES ====
        // KICK
        for p in dsp_kick.params_mut().params_mut() {
            //if p.name() == "Tuning" { p.set_init(46.0); } //46*1.74=80hz for audible range

            // Matching a sample:
            if p.name() == "Shell_Depth" { p.set_init(20.0); }
            if p.name() == "Reso_Head" { p.set_init(1.0); }
            if p.name() == "Tuning" { p.set_init(50.0); }
            if p.name() == "Detune_Range" { p.set_init(14.0); }
            if p.name() == "Reso_Tuning" { p.set_init(-3.0); }
            if p.name() == "Beater_Dampening" { p.set_init(50.0); }
            if p.name() == "Mix_Strike" { p.set_init(100.0); }
            //if p.name() == "Mix_Rattle" { p.set_init(20.0); }
            if p.name() == "Mix_Reso" { p.set_init(50.0); }
            if p.name() == "Decay" { p.set_init(0.3); }
            if p.name() == "Voice_Reverb" { p.set_init(0.0); }
            if p.name() == "Sample_Reverb" { p.set_init(0.0); }
            if p.name() == "Strike_Brightness" { p.set_init(25.0); }
            //if p.name() == "Strike_Pos" { p.set_init(5.0); }
            if p.name() == "Mix_Ring" { p.set_init(0.0); }
        }
        // SNARE
        for p in dsp_snare.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(-10.0); }
            if p.name() == "Shell_Depth" { p.set_init(6.0); } //5.5 - 6.6in deep (14in diameter)
            if p.name() == "Shell_Tuning" { p.set_init(340.0); }
            if p.name() == "Tuning" { p.set_init(180.0); }
            if p.name() == "Reso_Tuning" { p.set_init(3.0); }
            if p.name() == "Decay" { p.set_init(0.24); }
            if p.name() == "Beater_Dampening" { p.set_init(65.0); }
            if p.name() == "Mix_Reso" { p.set_init(100.0); }
            if p.name() == "Mix_Snare" { p.set_init(100.0); }
            if p.name() == "Mix_Ring" { p.set_init(100.0); }
            
        }
        // Rack Tom Sizes:
        // LOW TOM
        for p in dsp_lowtom.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(-20.0); }
            if p.name() == "Shell_Depth" { p.set_init(11.0); }
            if p.name() == "Reso_Head" { p.set_init(0.0); }
            if p.name() == "Tuning" { p.set_init(65.4); }
            //if p.name() == "Reso_Tuning" { p.set_init(0.4); }
            if p.name() == "Detune_Range" { p.set_init(12.0); }
            if p.name() == "Decay" { p.set_init(0.75); }
            if p.name() == "Beater_Dampening" { p.set_init(75.0); }
        }
        // MID TOM
        for p in dsp_midtom.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(5.0); }
            if p.name() == "Shell_Depth" { p.set_init(10.0); }
            if p.name() == "Reso_Head" { p.set_init(0.0); }
            if p.name() == "Tuning" { p.set_init(90.0); }
            //if p.name() == "Reso_Tuning" { p.set_init(0.4); }
            if p.name() == "Detune_Range" { p.set_init(10.0); }
            if p.name() == "Decay" { p.set_init(0.7); }
            if p.name() == "Beater_Dampening" { p.set_init(75.0); }
        }
        // HIGH TOM
        for p in dsp_hitom.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(30.0); }
            if p.name() == "Shell_Depth" { p.set_init(9.0); }
            if p.name() == "Reso_Head" { p.set_init(0.0); }
            if p.name() == "Tuning" { p.set_init(120.0); }
            //if p.name() == "Reso_Tuning" { p.set_init(0.4); }
            if p.name() == "Detune_Range" { p.set_init(8.0); }
            if p.name() == "Decay" { p.set_init(0.55); }
            if p.name() == "Beater_Dampening" { p.set_init(75.0); }
        }
        // HIHAT
        for p in dsp_hihat.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(10.0); }
            if p.name() == "Decay" { p.set_init(0.08); }
            if p.name() == "Voice_Punchiness" { p.set_init(0.0); }
            if p.name() == "Cymbal_Size" { p.set_init(14.0); }
            if p.name() == "Cymbal_Crash" { p.set_init(50.0); }
        }
        // CYMBAL
        for p in dsp_cymbal.params_mut().params_mut() {
            if p.name() == "Voice_Pan" { p.set_init(-10.0); }
            if p.name() == "Cymbal_Size" { p.set_init(20.0); }
            if p.name() == "Cymbal_Crash" { p.set_init(100.0); }
        }

        params.append_dsp(&mut dsp_output, "Global");
        let param_global_sensitivity = params.append_float(
            FloatParam::new(
                "Global Sensitivity",
                100.0,
                FloatRange::Linear {
                    min: 0.0,
                    max: 100.0,
                },
            )
            .with_unit("%")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Global",
        );
        let param_global_transpose = params.append_float(
            FloatParam::new(
                "Global Transpose",
                0.0,
                FloatRange::Linear {
                    min: -12.0,
                    max: 12.0,
                },
            )
            .with_unit("st")
            .with_value_to_string(Arc::new(|v| {
                format!("{value:.precision$}", precision = 1, value = v)
            })),
            "Global",
        );
        /*
        params.append_int(
            IntParam::new(
                "Global Bits",
                12,
                IntRange::Linear {
                    min: 8,
                    max: 16,
                },
            )
            .with_unit("bit")
            .with_value_to_string(Arc::new(|v| {
                match v {
                    16 => "Full ".to_string(),
                    _ => format!("{value}", value = v),
                }
            })),
            "Global",
        );
        */

        params.append_dsp(&mut dsp_kick, "Kick");
        params.append_dsp(&mut dsp_snare, "Snare");
        params.append_dsp(&mut dsp_lowtom, "Low Tom");
        params.append_dsp(&mut dsp_midtom, "Mid Tom");
        params.append_dsp(&mut dsp_hitom, "High Tom");
        params.append_dsp(&mut dsp_hihat, "Hihat");
        params.append_dsp(&mut dsp_cymbal, "Cymbal");

        let param_hihat_closed = dsp_hihat.param_index("Closed").unwrap();
        let param_snare_rimshot = dsp_snare.param_index("Strike_Rimshot").unwrap();

        params.append_field_int("tab_index", 0);
        params.append_field_string("selected_preset", "");


        Self {
            //params: Arc::new(OneTrickPluginParams::default()),
            params: Arc::new(params),
            //processor: FixedProcessor::default(),

            accum_buffer: ResizableBuffer::default(),

            dsp_output: RefCell::new(dsp_output),
            dsp_kick: RefCell::new(dsp_kick),
            dsp_snare: RefCell::new(dsp_snare),
            dsp_lowtom: RefCell::new(dsp_lowtom),
            dsp_midtom: RefCell::new(dsp_midtom),
            dsp_hitom: RefCell::new(dsp_hitom),
            dsp_hihat: RefCell::new(dsp_hihat),
            dsp_cymbal: RefCell::new(dsp_cymbal),

            param_global_sensitivity,
            param_global_transpose,
            param_hihat_closed,
            param_snare_rimshot,

            kick_indicator: Arc::new(AtomicBool::new(false)),
            snare_indicator: Arc::new(AtomicBool::new(false)),
            lowtom_indicator: Arc::new(AtomicBool::new(false)),
            midtom_indicator: Arc::new(AtomicBool::new(false)),
            hitom_indicator: Arc::new(AtomicBool::new(false)),
            hihat_indicator: Arc::new(AtomicBool::new(false)),
            cymbal_indicator: Arc::new(AtomicBool::new(false)),

            #[cfg(feature = "profile")]
            dsp_cost: Arc::new(AtomicI32::new(0)),

            sample_rate: 1, // Invalid until initilized
        }
    }
}

impl OneTrickUrchin {
    /// Set up fonts and default styles for Egui
    #[cfg(feature = "egui")]
    fn setup_egui_style(ctx: &Context) {
        // Start with the default fonts (we will be adding to them rather than replacing them).
        let mut fonts = FontDefinitions::default();

        BasicFonts::add_fonts(&mut fonts);
        Icons::add_fonts(&mut fonts);

        let mut style = style::Style::default();

        // Setup defaults for using generic controls:
        let font_id = FontId::new(17.0, FontFamily::Proportional);
        style.override_font_id = Some(font_id);
        //style.override_text_style = Some(TextStyle::Body);

        //style.widgets.noninteractive.bg_stroke = 
        //    Stroke::new(2.0, palette.grey().alpha(0.666).to_color32());

        // Set the base Style
        style.visuals = style::Visuals::dark();

        // Tooltips and Popup Windows:
        style.visuals.window_fill = Color32::DARK_GRAY;
        style.visuals.window_stroke = Stroke::default();
        style.visuals.popup_shadow = Shadow {offset: Vec2::new(0.0, 0.0), blur: 6.0, spread: 0.0, color: Color32::from_black_alpha(32)};
        style.visuals.override_text_color = Some(Color32::WHITE);
        
        style.visuals.selection.stroke = Stroke::new(2.0, Color32::WHITE);
        ctx.set_style(style);

        // Tell egui to use these fonts:
        ctx.set_fonts(fonts);
    }

    fn midi_channel_ok(channel: u8) -> bool {
        channel == 0 || channel == 9
    }

    fn map_midi_note(note: u8, _channel: u8) -> Option<GeneralMidiDrums> {
        GeneralMidiDrums::try_from(note).ok()
    }

    fn set_internal_params(&mut self) {
        // Gain Adjustments:
        self.dsp_kick.borrow_mut().set_param_by_name("GainAdjustment", 0.0);
        self.dsp_snare.borrow_mut().set_param_by_name("GainAdjustment", 0.0);
        self.dsp_hitom.borrow_mut().set_param_by_name("GainAdjustment", -5.5);
        self.dsp_midtom.borrow_mut().set_param_by_name("GainAdjustment", -5.5);
        self.dsp_lowtom.borrow_mut().set_param_by_name("GainAdjustment", -5.5);
        self.dsp_hihat.borrow_mut().set_param_by_name("GainAdjustment", 0.0);
        self.dsp_cymbal.borrow_mut().set_param_by_name("GainAdjustment", 0.0);
    }
}


impl Plugin for OneTrickUrchin {
    const NAME: &'static str = "OneTrick URCHIN";
    const VENDOR: &'static str = "Punk Labs";
    const URL: &'static str = env!("CARGO_PKG_HOMEPAGE");
    const EMAIL: &'static str = "contact@punklabs.com";

    const VERSION: &'static str = env!("CARGO_PKG_VERSION");

    const AUDIO_IO_LAYOUTS: &'static [AudioIOLayout] = &[
        #[cfg(feature = "multi_out")]
        AudioIOLayout {
            // Multi-Out Stereo
            main_input_channels: NonZeroU32::new(0),
            main_output_channels: NonZeroU32::new(NUM_CHANNELS),
            aux_input_ports: &[],
            // Two to five of these ports will be used at a time
            aux_output_ports: &[new_nonzero_u32(NUM_CHANNELS); AUX_BUS_COUNT],

            names: PortNames {
                layout: Some("Multi-Out Stereo"),

                main_input: None,
                main_output: Some("Output"),
                aux_inputs: &[],
                aux_outputs: &[
                    "Kick",
                    "Snare",
                    "Low Tom",
                    "Mid Tom",
                    "Hi Tom",
                    "Hihat",
                    "Cymbal",
                ],
            },
        },
        AudioIOLayout {
            // Stereo
            main_input_channels: NonZeroU32::new(0),
            main_output_channels: NonZeroU32::new(NUM_CHANNELS),
            names: PortNames {
                layout: Some("Stereo"),

                main_input: None,
                main_output: Some("Output"),
                aux_inputs: &[],
                aux_outputs: &[],
            },
            ..AudioIOLayout::const_default()
        },
    ];

    const MIDI_INPUT: MidiConfig = MidiConfig::Basic; //Basic or MidiCCs
    const MIDI_OUTPUT: MidiConfig = MidiConfig::None;

    const SAMPLE_ACCURATE_AUTOMATION: bool = false; // Too bad this doesn't split on Midi NoteOn events

    type SysExMessage = ();
    type BackgroundTask = ();

    fn params(&self) -> Arc<dyn Params> {
        self.params.clone()
    }

    fn initialize(
        &mut self,
        _audio_io_layout: &AudioIOLayout,
        buffer_config: &BufferConfig,
        _context: &mut impl InitContext<Self>,
    ) -> bool {
        // Resize buffers and perform other potentially expensive initialization operations here.
        // The `reset()` function is always called right after this function. You can remove this
        // function if you do not need it.

        let sample_rate = buffer_config.sample_rate as usize;

        //self.processor.initialize(48000, sample_rate, 2, MAX_BLOCK_SIZE);

        self.accum_buffer.resize(3, MAX_BLOCK_SIZE);

        //nih_log!("Sample Rate: {}", sample_rate);
        self.sample_rate = sample_rate;
        const SILENCE_MS: u32 = 100;
        for_each_dsp!(self, &mut |dsp: &mut dyn OneTrickDSPGeneral| {
            dsp.initialize(sample_rate)
                .resize_buffer(MAX_BLOCK_SIZE)
                .track_silence(SILENCE_MS);
        });

        self.set_internal_params();

        true
    }

    fn reset(&mut self) {
        // Reset buffers and envelopes here. This can be called from the audio thread and may not
        // allocate. You can remove this function if you do not need it.
        for_each_dsp!(self, &mut |dsp: &mut dyn OneTrickDSPGeneral| {
            dsp.reset();
        });
        
        self.set_internal_params();
    }

    fn process(
        &mut self,
        buffer: &mut Buffer,
        aux: &mut AuxiliaryBuffers,
        context: &mut impl ProcessContext<Self>,
    ) -> ProcessStatus {
        #[cfg(feature = "profile")]
        let dsp_timer = std::time::Instant::now();

        let _is_standalone = match context.plugin_api() {
            PluginApi::Standalone => true,
            _ => false,
        };

        ProcessEx::process_split_notes(
        //self.processor.process_split_notes(
            buffer,
            aux,
            MAX_BLOCK_SIZE,
            context,
            |event| {
                // MIDI
                match event {
                    NoteEvent::MidiPitchBend {
                        timing: _,
                        channel,
                        value,
                    } => {
                        if !Self::midi_channel_ok(channel) {
                            return;
                        }
                        // Pitch Shifting through synthesis:
                        // Artifact-free and Latency-free
                        self.dsp_kick.borrow_mut().pitch_wheel(value);
                        self.dsp_snare.borrow_mut().pitch_wheel(value);
                        self.dsp_lowtom.borrow_mut().pitch_wheel(value);
                        self.dsp_midtom.borrow_mut().pitch_wheel(value);
                        self.dsp_hitom.borrow_mut().pitch_wheel(value);
                        self.dsp_hihat.borrow_mut().pitch_wheel(value);
                        self.dsp_cymbal.borrow_mut().pitch_wheel(value);
                    }
                    NoteEvent::MidiCC {
                        timing: _,
                        channel,
                        cc,
                        value,
                    } => {
                        if !Self::midi_channel_ok(channel) {
                            return;
                        }
                        if cc == 1 {
                            self.dsp_output.borrow_mut().mod_wheel(value);
                        }
                    }
                    NoteEvent::NoteOn {
                        timing: _,
                        voice_id: _,
                        channel,
                        note,
                        velocity,
                    } => {
                        if !Self::midi_channel_ok(channel) {
                            return;
                        }
                        if let Some(midi_drum_note) = Self::map_midi_note(note, channel) {
                            let velocity = if let Some(p) =
                                self.params.param_float_at(self.param_global_sensitivity)
                            {
                                let sensitivity = p.value() * 0.01;
                                velocity * sensitivity + (1.0 - sensitivity) * (80.0 / 127.0)
                            } else {
                                velocity
                            };

                            match midi_drum_note {
                                GeneralMidiDrums::BassDrum | GeneralMidiDrums::AcousticBassDrum => {
                                    self.kick_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_kick.borrow_mut().trigger(velocity);

                                    // Cymbal Bleed/Chatter:
                                    // Would need some kind of TakeMaxVelocity in case fo a real cymbal hit
                                    //self.dsp_cymbal.borrow_mut().trigger(velocity*0.025);
                                    //self.dsp_cymbal.borrow_mut().set_param(self.param_hihat_closed, 0.0);
                                }
                                GeneralMidiDrums::AcousticSnare
                                | GeneralMidiDrums::ElectricSnare
                                | GeneralMidiDrums::SideStick => {
                                    self.snare_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    let is_rimshot = match midi_drum_note {
                                        GeneralMidiDrums::SideStick => 1.0,
                                        _ => 0.0
                                    };
                                    self.dsp_snare.borrow_mut().set_param(self.param_snare_rimshot, is_rimshot);
                                    self.dsp_snare.borrow_mut().trigger(velocity);
                                }
                                GeneralMidiDrums::LowFloorTom | GeneralMidiDrums::HighFloorTom => {
                                    self.lowtom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_lowtom.borrow_mut().trigger(velocity);
                                }
                                GeneralMidiDrums::LowTom | GeneralMidiDrums::LowMidTom => {
                                    self.midtom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_midtom.borrow_mut().trigger(velocity);
                                }
                                GeneralMidiDrums::HighMidTom | GeneralMidiDrums::HighTom => {
                                    self.hitom_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hitom.borrow_mut().trigger(velocity);
                                }
                                GeneralMidiDrums::ClosedHihat => {
                                    self.hihat_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hihat.borrow_mut().trigger(velocity);
                                    //self.dsp_hihat.borrow_mut().choke();
                                    self.dsp_hihat.borrow_mut().set_param(self.param_hihat_closed, 1.0);
                                }
                                GeneralMidiDrums::PedalHihat => {
                                    self.hihat_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hihat.borrow_mut().trigger(0.1);
                                    //self.dsp_hihat.borrow_mut().choke();
                                    self.dsp_hihat.borrow_mut().set_param(self.param_hihat_closed, 1.0);
                                }
                                GeneralMidiDrums::OpenHihat => {
                                    self.hihat_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_hihat.borrow_mut().trigger(velocity);
                                    //self.dsp_hihat.borrow_mut().choke();
                                    self.dsp_hihat.borrow_mut().set_param(self.param_hihat_closed, 0.0);
                                }
                                GeneralMidiDrums::CrashCymbal => {
                                    self.cymbal_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    self.dsp_cymbal.borrow_mut().trigger(velocity);
                                    self.dsp_cymbal.borrow_mut().set_param(self.param_hihat_closed, 0.0);
                                }
                                GeneralMidiDrums::RideCymbal => {
                                    self.cymbal_indicator.store(true, Ordering::Relaxed);
                                    self.dsp_output.borrow_mut().wake_up();
                                    //self.dsp_cymbal.borrow_mut().choke();
                                    self.dsp_cymbal.borrow_mut().set_param(self.param_hihat_closed, 1.0);
                                }
                                #[cfg(feature = "profile")]
                                GeneralMidiDrums::RideCymbal2 => {
                                    //B to profile Output DSP only
                                    self.dsp_output.borrow_mut().track_silence(1000); // Track for 1sec
                                    self.dsp_output.borrow_mut().wake_up();
                                }
                                _ => {}
                            }
                        }
                    }
                    _ => (),
                }
            },
            #[allow(unused_variables)]
            &mut |buffer: &mut Buffer, aux: &mut AuxiliaryBuffers, block_start, block_end| // AUDIO
            {
                //let frames = output[0].len() as i32;
                let frames = block_end - block_start;

                // Attempt to process audio in blocks split by events:
                let output = &mut buffer.slice2ch_range_mut(block_start, block_end);
                for channel in output.iter_mut() {
                    channel.fill(0.0);
                }

                //self.accum_buffer.clear();
                self.accum_buffer.buffer.clear_frames(frames);

                let pitch_shift = if let Some(p) =
                    self.params.param_float_at(self.param_global_transpose)
                {
                    p.value()
                } else {
                    0.0
                };

                for_each_voice_dsp!(self, &mut |dsp: &mut dyn OneTrickDSPGeneral| {
                    // Compute Voice:
                    if dsp.is_active() {
                        dsp.transpose(pitch_shift);
                        dsp.compute(frames, &[]);
                    } else {
                        dsp.skip_compute();
                    }
                    // Accumulate Voice:
                    dsp.add_to_buffer(&mut self.accum_buffer.buffer);
                });

                // Process Main Output:
                {
                    let mut dsp = self.dsp_output.borrow_mut();
                    if dsp.is_active() {
                        dsp.compute_to(frames, &self.accum_buffer.buffer.as_slice_actually_immutable(), Some(output));
                    } else {
                        dsp.skip_compute();
                    }
                }

                // Handle Aux Output:
                #[cfg(feature = "multi_out") ]
                {
                    // NIH Plug will zero out Aux buffers if it gets "weird data" from host...
                    // Ardour: Uses only the first Aux output..
                    let aux_supported: bool = !aux.outputs.is_empty()
                        && !aux.outputs[0].is_empty()
                        && !aux.outputs[0].as_slice()[0].is_empty();
                    if aux_supported {

                        let mut aux_outs = aux.outputs.iter_mut();

                        for_each_voice_dsp_grouped!(self, &mut |group: &[&mut dyn OneTrickDSPGeneral]| {
                            if let Some(aux_out) = aux_outs.next() {
                                if !aux_out.is_empty() && !aux_out.as_slice()[0].is_empty() {
                                    let mut is_first = true;
                                    for dsp in group {
                                        if dsp.compute_ran() {
                                            let mut slice = aux_out.slice2ch_range_mut(block_start, block_end);
                                            if is_first {
                                                dsp.write_to_slice(&mut slice);
                                                is_first = false;
                                            } else {
                                                dsp.add_to_slice(&mut slice);
                                            }
                                        }
                                    }
                                }
                            }
                        });
                    }
                }
            },
        );

        #[cfg(feature = "profile")]
        {
            let sec: f64 = dsp_timer.elapsed().as_secs_f64().try_into().unwrap();
            let max_sec: f64 = buffer.samples() as f64 / self.sample_rate as f64;
            let ratio_int = (sec / max_sec * 100. * 100000.) as i32; // Put it into %, then 5 decimals of precision
                                                                     // VU-style falling:
            let old_ratio_int = self.dsp_cost.load(Ordering::Relaxed);
            let ratio_int = ratio_int.max(old_ratio_int - 2000);
            self.dsp_cost.store(ratio_int, Ordering::Relaxed);
        }

        ProcessStatus::Normal
    }

    #[cfg(feature = "egui")]
    fn editor(&mut self, _async_executor: AsyncExecutor<Self>) -> Option<Box<dyn Editor>> {
        //async_executor.execute_gui((||{nih_log!("Hello from the GUI")})());
        //async_executor.execute_background((||{nih_log!("Hello from a Background Thread")})());

        let params = self.params.clone();
        let mut preset_manager = PresetManager::new("OneTrick URCHIN")
            .with_active_field(params.field_string("selected_preset"));

        let default_preset = Preset::from_param_defaults("Studio Drums", &params);
        preset_manager.add_factory(default_preset);
        preset_manager.add_factory_string("Clap Your Hands Say Nah", include_str!("assets/presets/Clap Your Hands Say Nah.preset"));
        preset_manager.add_factory_string("Spin That Kit", include_str!("assets/presets/Spin That Kit.preset"));
        preset_manager.add_factory_string("Disco Vinyl", include_str!("assets/presets/Disco Vinyl.preset"));
        preset_manager.add_factory_string("Boom Bap Vinyl", include_str!("assets/presets/Boom Bap Vinyl.preset"));
        preset_manager.add_factory_string("National Rhythm", include_str!("assets/presets/National Rhythm.preset"));
        preset_manager.add_factory_string("Woo Hoo", include_str!("assets/presets/Woo Hoo.preset"));
        preset_manager.add_factory_string("Total Trash Fire", include_str!("assets/presets/Total Trash Fire.preset"));
        preset_manager.add_factory_string("Ripper Kit", include_str!("assets/presets/Ripper Kit.preset"));
        preset_manager.add_factory_string("Sewer Surfer", include_str!("assets/presets/Sewer Surfer.preset"));
        preset_manager.add_factory_string("The 1980s", include_str!("assets/presets/The 1980s.preset"));
        preset_manager.add_factory_string("Blowout Kit", include_str!("assets/presets/Blowout Kit.preset"));
        preset_manager.add_factory_string("Bottles and Cans", include_str!("assets/presets/Bottles and Cans.preset"));
        preset_manager.add_factory_string("Catatonic", include_str!("assets/presets/Catatonic.preset"));
        preset_manager.add_factory_string("Summer Day", include_str!("assets/presets/Summer Day.preset"));
        preset_manager.add_factory_string("2 AM", include_str!("assets/presets/2 AM.preset"));
        preset_manager.add_factory_string("Electric Kit", include_str!("assets/presets/Electric Kit.preset"));
        preset_manager.add_factory_string("Little Bit", include_str!("assets/presets/Little Bit.preset"));
        preset_manager.add_factory_string("UK Garage", include_str!("assets/presets/UK Garage.preset"));
        preset_manager.add_factory_string("Phase 2", include_str!("assets/presets/Phase 2.preset"));
        

        preset_manager.refresh();

        //nih_log!("editor()");

        let tab_field: Arc<FieldI32> = params.field_int("tab_index").unwrap();

        /*
        let grunge1_svg: Arc<AtomicRefCell<Option<RetainedImage>>> =
            Arc::new(AtomicRefCell::new(None));
        let grunge1_svg_loader = grunge1_svg.clone();

        let grunge2_svg: Arc<AtomicRefCell<Option<RetainedImage>>> =
            Arc::new(AtomicRefCell::new(None));
        let grunge2_svg_loader = grunge2_svg.clone();
        */

        let palette = Palette::new(8)
            .shades(6)
            .white_level(1.0)
            .black_level(0.15)
            .saturation(0.75)
            .shift(0.49)
            .alt_hue_step(4.533)
            .dark_shift(0.1)
            .dark_desaturation(0.5);
        //let palette_alt = palette.alternate(1.5);

        let mut knob_style = ParamKnobStyle {
            radius: 35.0,
            indicator_style: ParamKnobIndicatorStyle::Continuous,
            ..Default::default()
        };
        let column_width = knob_style.required_width();


        let mut heading_style = LabelStyle::default_heading();
        heading_style.width = Some(column_width);
        heading_style.color = palette.black().into();
        heading_style.bg_color = palette.white().brightness(0.1).into(); //Color32::from_gray(128);
        let mut subheading_style = heading_style.clone();
        subheading_style.color = palette.black().into();
        subheading_style.bg_color = Color32::TRANSPARENT;
        subheading_style.shadow = None;
        heading_style.color = palette.white().into();



        let separator_style = SeparatorStyle {
            width: 2.0,
            color: palette.white().into(),
            ..Default::default()
        };
        knob_style.knob_color = palette.grey().into();
        knob_style.label_style.bg_color = palette.black().into();
        knob_style.label_style.color = palette.white().into();
        knob_style.label_style_hover.bg_color = palette.black().into();
        knob_style.label_style_hover.color = palette.color(0).into();
        knob_style.bg_color = palette.black().into();
        knob_style.indicator_fill_color = palette.grey().into();

        let mut knob_style_1 = knob_style.clone();
        knob_style_1.indicator_fill_color = palette.color_alt(0, 1).into();
        let mut knob_style_2 = knob_style.clone();
        knob_style_2.indicator_fill_color = palette.color(0).into();
        let mut knob_style_3 = knob_style.clone();
        knob_style_3.indicator_fill_color = palette.color(1).into();
        let mut knob_style_4 = knob_style.clone();
        knob_style_4.indicator_fill_color = palette.color(2).into();
        let mut knob_style_5 = knob_style.clone();
        knob_style_5.indicator_fill_color = palette.color(4).into();

        let mut slider_style = ParamSliderStyle {
            length: 70.0,
            ..Default::default()
        };
        /*
        slider_style.knob_aspect = 2.0/1.0;
        slider_style.knob_width = 20.0;
        slider_style.knob_round_ratio = 0.5;
        */
        slider_style.knob_color = knob_style.knob_color;
        slider_style.label_width = 70.0;
        slider_style.label_style.bg_color = knob_style.label_style.bg_color;
        slider_style.label_style.color = knob_style.label_style.color;
        slider_style.label_style_hover.bg_color = knob_style.label_style_hover.bg_color;
        slider_style.label_style_hover.color = knob_style.label_style.color;
        slider_style.bg_color = palette.black().into();
        slider_style.fill_color = Color32::LIGHT_GREEN;

        let mut slider_style_1 = slider_style.clone();
        slider_style_1.fill_color = knob_style_1.indicator_fill_color;
        let mut slider_style_2 = slider_style.clone();
        slider_style_2.fill_color = knob_style_2.indicator_fill_color;
        let mut slider_style_3 = slider_style.clone();
        slider_style_3.fill_color = knob_style_3.indicator_fill_color;
        let mut slider_style_4 = slider_style.clone();
        slider_style_4.fill_color = knob_style_4.indicator_fill_color;
        #[allow(clippy::redundant_clone)]
        let mut slider_style_5 = slider_style.clone();
        slider_style_5.fill_color = knob_style_5.indicator_fill_color;

        let show_credits = Arc::new(AtomicBool::new(false));
        //let anim_id = Id::new("local_anim"); // Can use in each context once
        let credits_anim_id = Id::new("credits_anim");
        let kick_anim_id = Id::new("kick_anim");
        let snare_anim_id = Id::new("snare_anim");
        let lowtom_anim_id = Id::new("lowtom_anim");
        let midtom_anim_id = Id::new("midtom_anim");
        let hitom_anim_id = Id::new("hitom_anim");
        let hihat_anim_id = Id::new("hihat_anim");
        let cymbal_anim_id = Id::new("cymbal_anim");

        #[cfg(feature = "profile")]
        let dsp_cost = self.dsp_cost.clone();

        let kick_indicator = self.kick_indicator.clone();
        let snare_indicator = self.snare_indicator.clone();
        let lowtom_indicator = self.lowtom_indicator.clone();
        let midtom_indicator = self.midtom_indicator.clone();
        let hitom_indicator = self.hitom_indicator.clone();
        let hihat_indicator = self.hihat_indicator.clone();
        let cymbal_indicator = self.cymbal_indicator.clone();

        create_egui_editor(
            self.params.editor_state.clone(),
            preset_manager,
            move |ctx, _| {
                // DPI:
                // egui_baseview needs to be updated to support DPI.
                // We can adjust egui's rendering, but not egui_baseview's resolution.
                //ctx.set_pixels_per_point(2.0);

                //nih_log!("Editor::Open()");
                Self::setup_egui_style(ctx);
            },
            move |ctx, setter, state| {
                #[cfg(feature = "stress_test")] // Add some randomizing functionality
                {
                    // Stress regular automation (not multiple at once!)
                    Preset::from_random_values(&params)
                        .filter_params_include(&["Global Volume"])
                        .apply_with_automation(&params, setter);

                    // Stress applying a preset
                    Preset::from_random_values(&params)
                        .apply(&params, setter);
                }

                //ctx.set_debug_on_hover(true);
                let preset_manager = state;

                //update()
                //let window_rect = ctx.available_rect();

                #[cfg(feature = "profile")]
                let dsp_cost = dsp_cost.load(Ordering::Relaxed) as f32 * 0.00001; // Scale for 5 decimals of precision

                let zoom_amount = 0.2;
                let zoom_time = 0.150;
                let indicator_raw = kick_indicator.swap(false, Ordering::Relaxed);
                let kick_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        kick_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = snare_indicator.swap(false, Ordering::Relaxed);
                let snare_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        snare_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = lowtom_indicator.swap(false, Ordering::Relaxed);
                let lowtom_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        lowtom_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = midtom_indicator.swap(false, Ordering::Relaxed);
                let midtom_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        midtom_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = hitom_indicator.swap(false, Ordering::Relaxed);
                let hitom_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        hitom_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = hihat_indicator.swap(false, Ordering::Relaxed);
                let hihat_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        hihat_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let indicator_raw = cymbal_indicator.swap(false, Ordering::Relaxed);
                let cymbal_indicator_zoom = 1.0
                    + ctx.animate_bool_with_time(
                        cymbal_anim_id,
                        indicator_raw,
                        if indicator_raw { 0.0 } else { zoom_time },
                    ) * zoom_amount;
                let show_credits_amount =
                    ctx.animate_bool_with_time(credits_anim_id, show_credits.load(Ordering::Relaxed), 0.300);

                if show_credits_amount > 0.0 {
                    Area::new("credits_area".into())
                        .fixed_pos(Pos2::new(0.0, 0.0))
                        .order(Order::Foreground)
                        .show(ctx, |ui| {
                            Frame::none()
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::same(10.0))
                                .fill(Color32::from_black_alpha(
                                    (200.0 * show_credits_amount) as u8,
                                ))
                                //.rounding(panel_rounding)
                                .show(ui, |ui| {
                                    let available_size =
                                        Vec2::new(ui.available_width(), ui.available_height());
                                    //let animated_offset = Vec2::new(0.0, available_size.y * (1.0-show_credits_amount));
                                    ui.painter().text(
                                        (available_size * 0.5).to_pos2(),
                                        Align2::CENTER_CENTER,
                                        format!(
                                            include_str!("CREDITS"),
                                            VERSION = env!("CARGO_PKG_VERSION")
                                        ),
                                        FontId::new(24.0, FontFamily::Name("Title".into())),
                                        Color32::from_white_alpha(
                                            (255.0 * show_credits_amount) as u8,
                                        ),
                                    );
                                    /*
                                    let galley = ui.painter().layout_no_wrap(
                                        include_str!("CREDITS").to_string(),
                                        FontId::new(20.0, FontFamily::Name("Title".into())),
                                        Color32::WHITE);
                                    let rect = Align2::CENTER_CENTER.anchor_rect(Rect::from_min_size((available_size * 0.5).to_pos2(), galley.size()));
                                    ui.painter().galley(rect.min, galley);
                                    */

                                    if ui
                                        .allocate_response(available_size, Sense::click())
                                        .clicked()
                                    {
                                        show_credits.store(false, Ordering::Relaxed);
                                    }
                                });
                        });
                }
                CentralPanel::default()
                    .frame(
                        Frame::none()
                            .outer_margin(Margin::same(0.0))
                            .inner_margin(Margin::same(10.0))
                            //.fill(panel_bg_color), //.rounding(Rounding{se: 20.0, sw: 20.0, ..Default::default()})
                            //.fill(palette.color(0).shade(1).into()),
                            .fill(palette.color(0).shade(1).into()),
                    )
                    .show(ctx, |ui| {
                        //ctx.set_debug_on_hover(false);

                        let header_height = 70.0;

                        ui.painter().rect_filled(
                            Rect::from_min_max(Pos2::new(0.0, 0.0), Pos2::new(415.0, header_height)),
                            Rounding::same(0.0),
                            palette.white().brightness(0.05).to_color32(),
                        );

                        ui.painter().parallelograms(
                            Rect::from_min_size(Pos2::new(325.0, 0.0), Vec2::new(220.0, header_height)),
                            1.0,
                            &[
                                //palette.color(0).shade(1).to_color32(),                                
                                palette.white().to_color32(),
                                palette.color(6).shade(1).to_color32(),
                                palette.color(5).shade(1).to_color32(),
                                palette.color(4).shade(1).to_color32(),
                                //palette.color(0).shade(1).to_color32(),
                            ],
                        );


                        Area::new("logo_area".into()) // Prevents layout being affected by Logo
                        .fixed_pos(Pos2::new(0.0, 0.0))
                        .order(Order::Foreground)
                        .show(ctx, |ui| {
                            #[cfg(feature = "preset_design")] // Add some randomizing functionality
                            {
                                if ui.put(
                                    Rect::from_min_size(Pos2::new(250.0, 18.0), Vec2::new(32.0, 32.0)),
                                    IconButton::new(Dingbat::Refresh)
                                    )
                                    .on_hover_text("Randomize current preset")
                                    .clicked()
                                {
                                    let preset = Preset::from_random_values(&params)
                                        .filter_params_exclude(&["Gain", "Volume", "Reso Head", "Reso Tuning", "Cutoff", "Pan", "Transpose", "Decay", "Mix", "Late"], &["Sample ToneGain"])
                                        .copy_params_by_prefix("Mid Tom", "Low Tom")
                                        .copy_params_by_prefix("Mid Tom", "High Tom");
                                    preset.apply(&params, setter);
                                }
                                if ui.put(
                                    Rect::from_min_size(Pos2::new(250.0+48.0, 18.0), Vec2::new(32.0, 32.0)),
                                    IconButton::new(Dingbat::TomDrumMedium)
                                    )
                                    .on_hover_text("Merge Tom Settings")
                                    .clicked()
                                {
                                    let preset = Preset::from_params("", &params)
                                        .filter_params_exclude(&["Shell Depth", "Gain", "Volume", "Reso Head", "Reso Tuning", "Cutoff", "Pan", "Transpose", "Decay", "Mix", "Late"], &["Sample ToneGain"])
                                        .copy_params_by_prefix("Mid Tom", "Low Tom")
                                        .copy_params_by_prefix("Mid Tom", "High Tom");
                                    preset.apply(&params, setter);
                                }
                            }

                            let response = ui.allocate_rect(
                                Rect::from_min_size(Pos2::new(0.0, 0.0),
                                Vec2::new(240.0, 70.0)),
                                Sense::click());
                            if response.clicked() {
                                show_credits.store(true, Ordering::Relaxed);
                            }
                            ui.painter().one_trick_logo(
                                "URCHIN",
                                Pos2::new(15.0, 35.0),
                                52.0,
                                palette.white().brightness(0.66).into(),
                                if response.hovered() {palette.color(5).into()} else {palette.white().into()},
                            );
                        });

                        ui.set_width(ui.available_width());

                        let panel_rounding = Rounding::same(15.0);
                        //let panel_shadow = Shadow::small_light();
                        let panel_shadow = Shadow {offset: Vec2::new(0.0, 0.0), blur: 24.0, spread: 0.0, color: Color32::from_black_alpha(24)};

                        // ============================================
                        // ============== MAIN 2 COLUMNS ==============
                        // ============================================
                        ui.horizontal(|ui| {
                            // ==========================================
                            // ============== VOICES PANEL ==============
                            // ==========================================
                            Frame::none()
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::symmetric(10.0, 10.0))
                                .fill(Color32::TRANSPARENT)
                                //.stroke(Stroke::new(1.0, palette.black()))
                                .show(ui, |ui| {
                                ui.vertical(|ui| { // Voices
                                    ui.set_width(75.0*8.0);
                                    ui.set_height(660.0);
                                    ui.add_space(header_height);
                                    /*
                                    ui.vertical_centered(|ui| {
                                        ui.add(
                                            Label::new("Kit").with_style(&heading_style),
                                        );
                                    });
                                    ui.add_space(5.0);
                                    */
                                    let tabs_style = IconTabsStyle{
                                        panel_rounding: 5.0.into(),
                                        panel_color: palette.white().brightness(0.25).alpha(0.8).into(),
                                        //bg_color_active: palette.color(3).into(),
                                        tab_color_active: palette.white().into(),
                                        color: palette.white().alpha(0.5).into(),
                                        color_active: palette.black().into(),
                                        ..Default::default()
                                    };

                                    let mut tab_index: u32 = tab_field.get() as u32;
                                    ui.add(IconTabs::new(
                                        &mut tab_index,
                                        tabs_style,
                                        vec![
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::KickDrum)
                                                .with_zoom(kick_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::SnareDrum)
                                                .with_zoom(snare_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::TomDrumLarge)
                                                .with_zoom(lowtom_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::TomDrumMedium)
                                                .with_zoom(midtom_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::TomDrumSmall)
                                                .with_zoom(hitom_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::Hihat)
                                                .with_zoom(hihat_indicator_zoom),
                                            IconTabStyle::default()
                                                .with_dingbat(Dingbat::Cymbal)
                                                .with_zoom(cymbal_indicator_zoom),
                                        ],
                                    ));
                                    tab_field.set(tab_index as i32);
                                    ui.add_space(15.0);
                                    let voices = [
                                        "Kick",
                                        "Snare",
                                        "Low Tom",
                                        "Mid Tom",
                                        "High Tom",
                                        "Hihat",
                                        "Cymbal",
                                    ];
                                    let voice_name = voices[tab_index as usize];
                                    let voice_type = if tab_index < 5 {
                                        0
                                    } else if tab_index < 7 {
                                        1
                                    } else {
                                        2
                                    };
                                    let voice_is_drum = voice_type==0;
                                    let _voice_is_snare = tab_index==1;
                                    let voice_is_cymbal = voice_type==1;
                                    let center_offset_x = 38.0;
                                    let spacing_y = 15.0;
                                    ui.horizontal(|ui| {
                                        let center_offset_x = 45.0;
                                        let justify_spacing_x = 22.0;
                                        ui.spacing_mut().item_spacing = vec2(justify_spacing_x, spacing_y);

                                        ui.set_height(232.0);
                                        ui.vertical(|ui| {
                                            ui.horizontal(|ui| {
                                                ui.add_space(center_offset_x);
                                                ui.add(Label::new("Mix").with_style(&subheading_style));
                                            });
                                            ui.horizontal(|ui| {
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!("{} Gain", voice_name)),
                                                        setter,
                                                    )
                                                    .with_label("Gain")
                                                    .with_style(&knob_style_5),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Reverb Send",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Reverb")
                                                    .with_style(&knob_style_5),
                                                );
                                            });
                                            ui.horizontal(|ui| {
                                                ui.add_space(center_offset_x-35.0);
                                                ui.add(
                                                    ParamSlider::for_param(
                                                        params.param_float(&format!("{} Pan", voice_name)),
                                                        setter,
                                                    )
                                                    .with_label("Pan")
                                                    .with_style(&slider_style_5)
                                                    .with_indicator_from_center()
                                                    .with_orientation(ParamSliderOrientation::Horizontal)
                                                    .with_length(140.0)
                                                    .with_label_width(70.0),
                                                );
                                            });
                                        });
                                        ui.add(Separator::new().with_style(&separator_style));
                                        ui.vertical(|ui| {
                                            ui.horizontal(|ui| {
                                                ui.add_space(center_offset_x);
                                                ui.add(Label::new("Filter").with_style(&subheading_style));
                                            });
                                            ui.horizontal(|ui| {
                                                ui.add_space(center_offset_x);
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Cutoff",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Lowpass")
                                                    .with_style(&knob_style_4),
                                                );
                                            });
                                            ui.horizontal(|ui| {
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Sample ToneFreq",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Tone Freq")
                                                    .with_style(&knob_style_4),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Sample ToneGain",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Tone Gain")
                                                    .with_style(&knob_style_4)
                                                    //.with_indicator_from_center(),
                                                );
                                            });
                                        });
                                        ui.add(Separator::new().with_style(&separator_style));
                                        ui.vertical(|ui| {
                                            ui.horizontal(|ui| {
                                                ui.add_space(center_offset_x);
                                                ui.add(Label::new("Chop").with_style(&subheading_style));
                                            });

                                            ui.horizontal(|ui| {
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Lateness",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Late")
                                                    .with_style(&knob_style_2),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Sample Length",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Slice")
                                                    .with_style(&knob_style_2),
                                                );
                                            });
                                            ui.horizontal(|ui| {
                                                //ui.add_space(center_offset_x);
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Voice Punchiness",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Punch")
                                                    .with_style(&knob_style_2),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Sample Speed",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Speed")
                                                    .with_style(&knob_style_2)
                                                    .with_indicator_from_center(),
                                                );
                                            });
                                        });
                                    });

                                    ui.add_space(20.0);
                                     ui.add(Separator::new()
                                         .with_style(&separator_style)
                                         .with_orientation(SeparatorOrientation::Horizontal));
                                    ui.add_space(20.0);
                                    

                                    ui.horizontal(|ui| {
                                        let justify_spacing_x = 8.0;
                                        ui.spacing_mut().item_spacing = vec2(justify_spacing_x, spacing_y);

                                        if voice_is_drum {
                                            ui.vertical(|ui| {
                                                ui.add(Label::new("Strike").with_style(&subheading_style));
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Mix Strike",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Amount")
                                                    .with_style(&knob_style_3),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Strike Brightness",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Tone")
                                                    .with_style(&knob_style_3),
                                                );
                                            });

                                            ui.add(Separator::new().with_style(&separator_style));

                                            ui.vertical(|ui| {
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(Label::new("Drum").with_style(&subheading_style));
                                                });
    
                                                ui.horizontal(|ui| {
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Transpose",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Tune")
                                                        .with_style(&knob_style_1),
                                                    );
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Beater Damp",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Damp")
                                                        .with_style(&knob_style_1),
                                                    );
                                                });
                                                ui.horizontal(|ui| {
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Decay",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Decay")
                                                        .with_style(&knob_style_1),
                                                    );
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Detune Range",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Detune")
                                                        .with_style(&knob_style_1),
                                                    );
                                                });
                                            });

                                            ui.add(Separator::new().with_style(&separator_style));

                                            ui.vertical(|ui| {
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(Label::new("Shell").with_style(&subheading_style));
                                                });
    
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Shell Depth",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Size")
                                                        .with_style(&knob_style_4)
                                                        .with_indicator_style(ParamKnobIndicatorStyle::Dots)
                                                        .with_indicator_count(20),
                                                    );
                                                });
                                                ui.horizontal(|ui| {
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Shell Damp",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Damp")
                                                        .with_style(&knob_style_4),
                                                    );
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Mix Ring",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Ring")
                                                        .with_style(&knob_style_4),
                                                    );
                                                });
                                            });

                                            ui.add(Separator::new().with_style(&separator_style));

                                            ui.vertical(|ui| {
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(Label::new("Reso").with_style(&subheading_style));
                                                });
   
                                                ui.add_space(30.0);
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(
                                                        ParamSlider::for_param(
                                                            params.param_bool(&format!(
                                                                "{} Reso Head",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Toggle")
                                                        .with_style(&slider_style_5)
                                                        .with_label_only_value()
                                                        .with_length(40.0)
                                                        .with_tick_count(2)
                                                        .without_fill(),
                                                    );
                                                });
                                                ui.horizontal(|ui| {
                                                    let reso_head = if let Some(param) = params.param_bool(&format!(
                                                        "{} Reso Head",
                                                        voice_name
                                                    )) {
                                                        param.value()
                                                    } else {
                                                        false
                                                    };
                                                    ui.set_enabled(reso_head);
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Reso Tuning",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Tune")
                                                        .with_style(&knob_style_5),
                                                    );
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Mix Reso",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Mix")
                                                        .with_style(&knob_style_5),
                                                    );
                                                });
                                            });
                                        }
                                        if voice_is_cymbal {
                                            ui.add_space((ui.available_width()-200.0)/2.0 - 20.0);
                                            ui.vertical(|ui| {
                                                ui.add(Label::new("Strike").with_style(&subheading_style));
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Mix Strike",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Amount")
                                                    .with_style(&knob_style_3),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float(&format!(
                                                            "{} Strike Brightness",
                                                            voice_name
                                                        )),
                                                        setter,
                                                    )
                                                    .with_label("Tone")
                                                    .with_style(&knob_style_3),
                                                );
                                            });

                                            ui.add(Separator::new().with_style(&separator_style).with_padding(20.0));

                                            ui.vertical(|ui| {
                                                ui.horizontal(|ui|{
                                                    ui.add_space(center_offset_x);
                                                    ui.add(Label::new("Cymbal").with_style(&subheading_style));
                                                });
                                                ui.horizontal(|ui| {
                                                    ui.add_space(center_offset_x);
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Size",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Size")
                                                        .with_style(&knob_style_4),
                                                    );
                                                });
                                                ui.horizontal(|ui|{
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Damp",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Damp")
                                                        .with_style(&knob_style_4),
                                                    );
                                                    ui.add(
                                                        ParamKnob::for_param(
                                                            params.param_float(&format!(
                                                                "{} Crash",
                                                                voice_name
                                                            )),
                                                            setter,
                                                        )
                                                        .with_label("Crash")
                                                        .with_style(&knob_style_4),
                                                    );
                                                });
                                        });
                                        }
                                    });
                                    // ===============================================
                                    // ===============================================
                                    // ===============================================
                                    // ===============================================
                                    // ===============================================
                                    #[cfg(feature = "preset_design")] // Add some randomizing functionality
                                    {
                                        if ui.put(
                                            Rect::from_min_size(ui.next_widget_position() + Vec2::new(0.0, -300.0), Vec2::new(32.0, 32.0)),
                                            IconButton::new(Dingbat::Refresh)
                                            )
                                            .on_hover_text("Randomize Voice")
                                            .clicked()
                                        {
                                            let preset = Preset::from_random_values(&params)
                                                .filter_params_include(&[voice_name])
                                                .filter_params_exclude(&["Gain", "Volume", "Reso Head", "Reso Tuning", "Cutoff", "Pan", "Transpose", "Decay", "Mix", "Late"], &["Sample ToneGain"]);
                                            preset.apply(&params, setter);
                                        }
                                    }
                                });
                            });
                            
                            //ui.add(Separator::new().with_style(&separator_style));

                            // ============================================
                            // ============== GLOBALS PANELS ==============
                            // ============================================
                            Frame::none() // Global Panel Wrapper
                                /*
                                .outer_margin(Margin::same(0.0))
                                .inner_margin(Margin::same(20.0))
                                .rounding(panel_rounding)
                                .fill(palette.white().into())
                                .stroke(Stroke::new(2.0, palette.black()))
                                */
                                .show(ui, |ui| {

                                let frame_outer_margin = Margin {
                                    top:5.0,
                                    ..Default::default()
                                };
                                //let frame_inner_margin = Margin::same(10.0);
                                let frame_inner_margin = Margin{left: 0.0, right: 0.0, top: 5.0, bottom: 15.0};
                                let subframe_inner_margin = Margin::same(10.0);
                                                                
                                ui.set_max_width(75.0*5.0+30.0);
                                ui.vertical_centered(|ui| {
                                    // ==========================================
                                    // ============== REVERB PANEL ==============
                                    // ==========================================
                                    ui.add_space(5.0);
                                    Frame::none()
                                        .outer_margin(Margin::same(0.0))
                                        .inner_margin(frame_inner_margin)
                                        .fill(heading_style.bg_color)
                                        .rounding(panel_rounding)
                                        .shadow(panel_shadow)
                                        .show(ui, |ui| {
                                        ui.vertical_centered(|ui| {
                                            ui.add(
                                                Label::new("Track Sim")
                                                .with_style(&heading_style)
                                            );
                                        });
                                        Frame::none()
                                        .outer_margin(Margin::same(0.0))
                                        .inner_margin(subframe_inner_margin)
                                        .fill(palette.color(3).shade(1).into())
                                        .show(ui, |ui| {
                                        ui.vertical_centered(|ui| {
                                            ui.add(
                                                Label::new("Studio").with_style(&subheading_style),
                                            );
                                            ui.add_space(5.0);
                                            ui.horizontal(|ui| {
                                                //ui.add_space(42.0); //Center... there's probably a simpler way.. *shrug*
                                                /*
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb Time"),
                                                        setter,
                                                    )
                                                    .with_label("Time")
                                                    .with_style(&knob_style),
                                                );
                                                */
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb Size"),
                                                        setter,
                                                    )
                                                    .with_label("Size")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb Diffusion"),
                                                        setter,
                                                    )
                                                    .with_label("Diff")
                                                    .with_style(&knob_style)
                                                    .with_indicator_style(ParamKnobIndicatorStyle::Dots),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb Damp"),
                                                        setter,
                                                    )
                                                    .with_label("Damp")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb Highpass"),
                                                        setter,
                                                    )
                                                    .with_label("Highpass")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Reverb"),
                                                        setter,
                                                    )
                                                    .with_label("Amount")
                                                    .with_style(&knob_style_5),
                                                );
                                                #[cfg(feature = "preset_design")] // Add some randomizing functionality
                                                {
                                                    if ui.put(
                                                        Rect::from_min_size(ui.next_widget_position() + Vec2::new(-38.0, -78.0), Vec2::new(32.0, 32.0)),
                                                        IconButton::new(Dingbat::Refresh)
                                                        )
                                                        .on_hover_text("Randomize Reverb")
                                                        .clicked()
                                                    {
                                                        let preset = Preset::from_random_values(&params)
                                                            .filter_params_include(&["Global Reverb"]);
                                                        preset.apply(&params, setter);
                                                    }
                                                }
                                                
                                            });
                                        });
                                        });
                                        Frame::none()
                                        .outer_margin(Margin::same(0.0))
                                        .inner_margin(subframe_inner_margin)
                                        .fill(palette.color(4).shade(1).into())
                                        .show(ui, |ui| {
                                        // =========================================
                                        // ============== MEDIA PANEL ==============
                                        // =========================================
                                        ui.vertical_centered(|ui| {
                                            ui.add(
                                                Label::new("Media").with_style(&subheading_style),
                                            );
                                            ui.add_space(5.0);
                                            ui.horizontal(|ui| {
                                                //ui.add_space(42.0); //Center... there's probably a simpler way.. *shrug*
                                                ui.add(
                                                    ParamSlider::for_param(
                                                        params.param_int("Global Noise Type"),
                                                        setter,
                                                    )
                                                    .with_label("Format")
                                                    .with_style(&slider_style)
                                                    .with_label_only_value()
                                                    .with_tick_count(2)
                                                    .without_fill(),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Noise Amount"),
                                                        setter,
                                                    )
                                                    .with_label("Noise")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Highpass"),
                                                        setter,
                                                    )
                                                    .with_label("Low Cut")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Lowpass"),
                                                        setter,
                                                    )
                                                    .with_label("High Cut")
                                                    .with_style(&knob_style)
                                                    .with_indicator_from_end(),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Saturation"),
                                                        setter,
                                                    )
                                                    .with_label("Saturation")
                                                    .with_style(&knob_style),
                                                );
                                                #[cfg(feature = "preset_design")] // Add some randomizing functionality
                                                {
                                                    if ui.put(
                                                        Rect::from_min_size(ui.next_widget_position() + Vec2::new(-38.0, -78.0), Vec2::new(32.0, 32.0)),
                                                        IconButton::new(Dingbat::Refresh)
                                                        )
                                                        .on_hover_text("Randomize Media")
                                                        .clicked()
                                                    {
                                                        let preset = Preset::from_random_values(&params)
                                                            .filter_params_include(&["Global Noise", "Global Highpass", "Global Lowpass", "Global Saturation"]);
                                                        preset.apply(&params, setter);
                                                    }
                                                }
                                            });
                                        });
                                        });
                                        Frame::none()
                                        .outer_margin(Margin::same(0.0))
                                        .inner_margin(subframe_inner_margin)
                                        .fill(palette.color(5).shade(1).into())
                                        .show(ui, |ui| {
                                        // ============================================
                                        // ============== SAMPLING PANEL ==============
                                        // ============================================
                                        ui.vertical_centered(|ui| {
                                            ui.add(
                                                Label::new("Sampler").with_style(&subheading_style),
                                            );
                                            ui.add_space(5.0);
                                            ui.horizontal(|ui| {
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Transpose"),
                                                        setter,
                                                    )
                                                    .with_label("Speed")
                                                    .with_style(&knob_style)
                                                    .with_indicator_from_center(),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Samplerate"),
                                                        setter,
                                                    )
                                                    .with_label("Samplerate")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_int("Global Bits"),
                                                        setter,
                                                    )
                                                    .with_label("Bits")
                                                    .with_style(&knob_style)
                                                    .with_indicator_style(ParamKnobIndicatorStyle::Dots)
                                                    .with_indicator_count(13),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Sensitivity"),
                                                        setter,
                                                    )
                                                    .with_label("Sensitivity")
                                                    .with_style(&knob_style),
                                                );
                                                ui.add(
                                                    ParamKnob::for_param(
                                                        params.param_float("Global Volume"),
                                                        setter,
                                                    )
                                                    .with_label("Volume")
                                                    .with_style(&knob_style_5),
                                                );
                                                #[cfg(feature = "preset_design")] // Add some randomizing functionality
                                                {
                                                    if ui.put(
                                                        Rect::from_min_size(ui.next_widget_position() + Vec2::new(-38.0, -78.0), Vec2::new(32.0, 32.0)),
                                                        IconButton::new(Dingbat::Refresh)
                                                        )
                                                        .on_hover_text("Randomize Sampler")
                                                        .clicked()
                                                    {
                                                        let preset = Preset::from_random_values(&params)
                                                            .filter_params_include(&["Global Samplerate", "Global Bits", "Global Sensitivity"]);
                                                        preset.apply(&params, setter);
                                                    }
                                                }
                                            });
                                        });
                                        });
                                    });
                                    // ===========================================
                                    // ============== PRESETS PANEL ==============
                                    // ===========================================
                                    Frame::none()
                                        .outer_margin(frame_outer_margin)
                                        .inner_margin(frame_inner_margin)
                                        //.fill(palette.color(6).shade(1).into())
                                        .fill(heading_style.bg_color)
                                        .rounding(panel_rounding)
                                        .shadow(panel_shadow)
                                        .show(ui, |ui| {

                                        ui.vertical_centered(|ui| {
                                            ui.add(
                                                Label::new("Presets")
                                                .with_style(&heading_style)
                                            );
                                        });

                                        Frame::none()
                                        .outer_margin(Margin::same(0.0))
                                        .inner_margin(subframe_inner_margin)
                                        .fill(palette.color(6).shade(1).into())
                                        .show(ui, |ui| {
                                            // ============================================
                                            // ============== PRESETS PANEL ==============
                                            // ============================================
                                            //let preset_manager: PresetManager = preset_manager.borrow();
                                            let list_style = PresetListStyle {
                                                bg_color: None,//Some(palette.dark_layer(1).into()),
                                                color: palette.black().into(),
                                                icon_style: IconButtonStyle{
                                                    color: palette.black().into(),
                                                    bg_color: Color32::TRANSPARENT,
                                                    color_hover: palette.white().into(),
                                                    bg_color_hover: palette.black().into(),
                                                    ..Default::default()
                                                },
                                                label_style: LabelStyle{
                                                    bg_color: palette.black().alpha(0.25).into(),
                                                    bg_color_hover: Some(palette.black().alpha(0.5).into()),
                                                    color: palette.white().brightness(0.95).into(),
                                                    color_hover: Some(palette.white().into()),
                                                    ..Default::default()
                                                },
                                                highlight_color: palette.color(0).into(),
                                                highlight_bg_color: palette.black().alpha(0.9).into(),
                                                search_bg_color: Some(palette.black().alpha(0.5).into()),
                                                search_color: Some(palette.white().into()),
                                                popup_offset_normalized: vec2(-1.0, 0.0),
                                                ..Default::default()
                                            };

                                            ui.vertical_centered(|ui| {
                                                ui.set_width(ui.available_width());
                                                ui.set_height(ui.available_height());
                                                PresetList::new()
                                                    .with_style(&list_style)
                                                    .show(ui, preset_manager, &params, setter);
                                                /*
                                                ListView::new([].iter(), ())
                                                    .title("Search".into())
                                                    .hold_text("something".into())
                                                    .striped()
                                                    .show(ctx, ui);
                                                */
                                            });
                                        });
                                    });
                                });
                            });
                        });
                    });
            },
        )
    }
}

impl ClapPlugin for OneTrickUrchin {
    const CLAP_ID: &'static str = "com.punklabs.onetrick.urchin";
    const CLAP_DESCRIPTION: Option<&'static str> = Some("A Drum Machine");
    const CLAP_MANUAL_URL: Option<&'static str> = Some(Self::URL);
    const CLAP_SUPPORT_URL: Option<&'static str> = None;
    const CLAP_FEATURES: &'static [ClapFeature] = &[
        ClapFeature::Instrument,  //Plugin Category
        ClapFeature::DrumMachine, // Plugin Sub-Category
        ClapFeature::Stereo,
    ]; // Audio Capabilities
       /*
       const CLAP_POLY_MODULATION_CONFIG: Option<PolyModulationConfig> = Some(PolyModulationConfig {
           // If the plugin's voice capacity changes at runtime (for instance, when switching to a
           // monophonic mode), then the plugin should inform the host in the `initialize()` function
           // as well as in the `process()` function if it changes at runtime using
           // `context.set_current_voice_capacity()`
           max_voice_capacity: NUM_VOICES,
           // This enables voice stacking in Bitwig.
           supports_overlapping_voices: true,
       });
       */
}

impl Vst3Plugin for OneTrickUrchin {
    const VST3_CLASS_ID: [u8; 16] = *b"OneTrick-URCHIN ";
    //const VST3_CATEGORIES: &'static str = "Instrument|Synth";
    const VST3_SUBCATEGORIES: &'static [Vst3SubCategory] =
        &[Vst3SubCategory::Instrument, Vst3SubCategory::Synth];
}

nih_export_clap!(OneTrickUrchin);
nih_export_vst3!(OneTrickUrchin);
